@@ -1,6 +1,24 @@
-/* pstree.c - display process tree */
+/*
+ * pstree.c - display process tree
+ *
+ * Copyright (C) 1993-2002 Werner Almesberger
+ * Copyright (C) 2002-2005 Craig Small
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with this program; if not, write to the Free Software
+ *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ */
 
-/* Copyright 1993-1999 Werner Almesberger. See file COPYING for details. */
 
 #ifdef HAVE_CONFIG_H
 #include <config.h>
@@ -26,9 +44,11 @@
 #include "i18n.h"
 #include "comm.h"
 
-#ifdef FLASK_LINUX
-#include <fs_secure.h>
-#endif /*FLASK_LINUX*/
+#ifdef WITH_SELINUX
+#include <selinux/selinux.h>
+#endif /*WITH_SELINUX*/
+
+extern const char *__progname;
 
 #ifndef MAX_DEPTH
 #define MAX_DEPTH    100
@@ -57,9 +77,9 @@ typedef struct _proc
   int argc;			/* with -a   : number of arguments, -1 if swapped    */
   pid_t pid;
   uid_t uid;
-#ifdef FLASK_LINUX
-  security_id_t sid;
-#endif /*FLASK_LINUX*/
+#ifdef WITH_SELINUX
+  security_context_t scontext;
+#endif /*WITH_SELINUX*/
   int highlight;
   struct _child *children;
   struct _proc *parent;
@@ -107,10 +127,9 @@ static PROC *list = NULL;
 static int width[MAX_DEPTH], more[MAX_DEPTH];
 static int print_args = 0, compact = 1, user_change = 0, pids = 0, by_pid = 0,
   trunc = 1, wait_end = 0;
-#ifdef FLASK_LINUX
-static int show_sids    = 0;
+#ifdef WITH_SELINUX
 static int show_scontext = 0;
-#endif /*FLASK_LINUX*/
+#endif /*WITH_SELINUX*/
 static int output_width = 132;
 static int cur_x = 1;
 static char last_char = 0;
@@ -160,38 +179,16 @@ out_int (int x)			/* non-negative integers only */
   return digits;
 }
 
-#ifdef FLASK_LINUX
+#ifdef WITH_SELINUX
 static void 
-out_sid ( security_id_t sid )
+out_scontext ( security_context_t scontext )
 {
-  if ( (int) sid >= 0 )
-    out_int((int) sid);
-  else
-    out_string("??");
-}
-
-static void 
-out_scontext ( security_id_t sid )
-{
-  static char buf[256];
-  int security_sid_to_context();
-  int len = sizeof(buf);
-  int rv;
-
-  bzero(buf,256);
-
-  rv = security_sid_to_context((int)sid, buf, &len);
-  if ( rv ) {
-    out_string("`??\'"); /* punt */
-  }
-  else {
     out_string("`");
-    out_string(buf);
-    out_string("\'");
-  }
+    out_string(scontext);
+    out_string("'");
 }
-#endif /*FLASK_LINUX*/
-
+#endif /*WITH_SELINUX*/
+  
 
 static void
 out_newline (void)
@@ -215,13 +212,13 @@ find_proc (pid_t pid)
   return walk;
 }
 
-#ifdef FLASK_LINUX
+#ifdef WITH_SELINUX
 static PROC *
-new_proc(const char *comm, pid_t pid, uid_t uid, security_id_t sid)
-#else  /*FLASK_LINUX*/
+new_proc(const char *comm, pid_t pid, uid_t uid, security_context_t scontext)
+#else  /*WITH_SELINUX*/
 static PROC *
 new_proc (const char *comm, pid_t pid, uid_t uid)
-#endif /*FLASK_LINUX*/
+#endif /*WITH_SELINUX*/
 {
   PROC *new;
 
@@ -234,9 +231,9 @@ new_proc (const char *comm, pid_t pid, uid_t uid)
   new->pid = pid;
   new->uid = uid;
   new->highlight = 0;
-#ifdef FLASK_LINUX
-  new->sid = sid;
-#endif /*FLASK_LINUX*/
+#ifdef WITH_SELINUX
+  new->scontext = scontext;
+#endif /*WITH_SELINUX*/
   new->children = NULL;
   new->parent = NULL;
   new->next = list;
@@ -305,24 +302,24 @@ set_args (PROC * this, const char *args, int size)
     this->argv[i] = start = strchr (start, 0) + 1;
 }
 
-#ifdef FLASK_LINUX
+#ifdef WITH_SELINUX
 static void
 add_proc(const char *comm, pid_t pid, pid_t ppid, uid_t uid,
-         const char *args, int size, security_id_t sid)
-#else  /*FLASK_LINUX*/
+         const char *args, int size, security_context_t scontext)
+#else  /*WITH_SELINUX*/
 static void
 add_proc (const char *comm, pid_t pid, pid_t ppid, uid_t uid,
 	  const char *args, int size)
-#endif /*FLASK_LINUX*/
+#endif /*WITH_SELINUX*/
 {
   PROC *this, *parent;
 
   if (!(this = find_proc (pid)))
-#ifdef FLASK_LINUX
-    this = new_proc(comm, pid, uid, sid);
-#else  /*FLASK_LINUX*/
+#ifdef WITH_SELINUX
+    this = new_proc(comm, pid, uid, scontext);
+#else  /*WITH_SELINUX*/
     this = new_proc (comm, pid, uid);
-#endif /*FLASK_LINUX*/
+#endif /*WITH_SELINUX*/
   else
     {
       strcpy (this->comm, comm);
@@ -333,11 +330,11 @@ add_proc (const char *comm, pid_t pid, pid_t ppid, uid_t uid,
   if (pid == ppid)
     ppid = 0;
   if (!(parent = find_proc (ppid)))
-#ifdef FLASK_LINUX
-    parent = new_proc("?", ppid, 0, sid);
-#else  /*FLASK_LINUX*/
+#ifdef WITH_SELINUX
+    parent = new_proc("?", ppid, 0, scontext);
+#else  /*WITH_SELINUX*/
     parent = new_proc ("?", ppid, 0);
-#endif /*FLASK_LINUX*/
+#endif /*WITH_SELINUX*/
   add_child (parent, this);
   this->parent = parent;
 }
@@ -429,25 +426,17 @@ dump_tree (PROC * current, int level, int rep, int leaf, int last,
       else
 	(void) out_int (current->uid);
     }
-#ifdef FLASK_LINUX
-  if ( show_sids ) {
-    out_char (info++ ? ',' : '(');
-    out_sid(current->sid);
-  }
+#ifdef WITH_SELINUX
   if ( show_scontext ) {
     out_char (info++ ? ',' : '(');
-    out_scontext(current->sid);
+    out_scontext(current->scontext);
   }
-#endif /*FLASK_LINUX*/
+#endif /*WITH_SELINUX*/
   if ((swapped && print_args && current->argc < 0) || (!swapped && info))
     out_char (')');
   if (current->highlight && (tmp = tgetstr ("me", NULL)))
     tputs (tmp, 1, putchar);
-#ifdef FLASK_LINUX
-  if (show_scontext || print_args)
-#else  /*FLASK_LINUX*/
   if (print_args)
-#endif /*FLASK_LINUX*/
     {
       for (i = 0; i < current->argc; i++)
 	{
@@ -472,20 +461,20 @@ dump_tree (PROC * current, int level, int rep, int leaf, int last,
 	    }
 	}
     }
-#ifdef FLASK_LINUX
+#ifdef WITH_SELINUX
   if ( show_scontext || print_args || ! current->children )
-#else  /*FLASK_LINUX*/
+#else  /*WITH_SELINUX*/
   if (print_args || !current->children)
-#endif /*FLASK_LINUX*/
+#endif /*WITH_SELINUX*/
     {
       while (closing--)
 	out_char (']');
       out_newline ();
-#ifdef FLASK_LINUX
+#ifdef WITH_SELINUX
       if ( show_scontext || print_args )
-#else /*FLASK_LINUX*/
+#else /*WITH_SELINUX*/
       if (print_args)
-#endif /*FLASK_LINUX*/
+#endif /*WITH_SELINUX*/
 	{
 	  more[level] = !last;
 	  width[level] = swapped + (comm_len > 1 ? 0 : -1);
@@ -544,6 +533,9 @@ dump_by_user (PROC * current, uid_t uid)
 {
   const CHILD *walk;
 
+  if (!current)
+    return;
+
   if (current->uid == uid)
     {
       if (dumped)
@@ -575,9 +567,10 @@ read_proc (void)
   pid_t pid, ppid;
   int fd, size;
   int empty;
-#ifdef FLASK_LINUX
-  security_id_t sid = -1;
-#endif /*FLASK_LINUX*/
+#ifdef WITH_SELINUX
+  security_context_t scontext = NULL;
+  int selinux_enabled=is_selinux_enabled()>0;
+#endif /*WITH_SELINUX*/
 
   if (!print_args)
     buffer = NULL;
@@ -602,11 +595,15 @@ read_proc (void)
 	  {
 	    empty = 0;
 	    sprintf (path, "%s/%d", PROC_BASE, pid);
-#ifdef FLASK_LINUX
-            if (fstat_secure(fileno(file),&st,&sid) < 0)
-#else /*FLASK_LINUX*/
+#ifdef WITH_SELINUX
+	    if (selinux_enabled)
+	      if (getpidcon(pid,&scontext) < 0)
+		{
+		  perror (path);
+		  exit (1);
+		}
+#endif /*WITH_SELINUX*/
             if (stat (path, &st) < 0)
-#endif /*FLASK_LINUX*/
 	    {
 		perror (path);
 		exit (1);
@@ -616,11 +613,10 @@ read_proc (void)
             {
               memset(comm, '\0', COMM_LEN+1);
               tmpptr = strrchr(readbuf, ')'); /* find last ) */
-              *tmpptr = '\0';
               /* We now have readbuf with pid and cmd, and tmpptr+2
                * with the rest */
               /*printf("readbuf: %s\n", readbuf);*/
-              if (sscanf(readbuf, "%*d (%15c", comm) == 1)
+              if (sscanf(readbuf, "%*d (%15[^)]", comm) == 1)
               {
                 /*printf("tmpptr: %s\n", tmpptr+2);*/
                 if (sscanf(tmpptr+2, "%*c %d", &ppid) == 1)
@@ -630,12 +626,47 @@ read_proc (void)
 		(file, "%d (%s) %c %d", &dummy, comm, (char *) &dummy,
 		 &ppid) == 4)
  */
+		{
+		   DIR *taskdir;
+		   struct dirent *dt;
+		   char *taskpath;
+		   char *threadname;
+		   int thread;
+		   
+		   if (!(taskpath = malloc(strlen(path) + 10))) {
+		      exit (2);
+		   }
+		   sprintf (taskpath, "%s/task", path);
+		   
+		   if ((taskdir=opendir(taskpath))!=0) {
+		      /* if we have this dir, we're on 2.6 */
+		      if (!(threadname = malloc(strlen(comm) + 3))) {
+			 exit (2);
+		      }
+		      sprintf(threadname,"{%s}",comm);
+		      while ((dt = readdir(taskdir)) != NULL) {
+			 if ((thread=atoi(dt->d_name)) !=0) {
+			    if (thread != pid) {
+#ifdef WITH_SELINUX
+			       add_proc(threadname, thread, pid, st.st_uid, NULL, 0, scontext);
+#else  /*WITH_SELINUX*/
+			       add_proc(threadname, thread, pid, st.st_uid, NULL, 0);
+#endif /*WITH_SELINUX*/
+			    }
+			 }
+		      }
+		      free(threadname);
+		      (void) closedir(taskdir);
+		   }
+		   free(taskpath);
+		}
+
 		if (!print_args)
-#ifdef FLASK_LINUX
-		  add_proc(comm, pid, ppid, st.st_uid, NULL, 0, sid);
-#else  /*FLASK_LINUX*/
+#ifdef WITH_SELINUX
+		  add_proc(comm, pid, ppid, st.st_uid, NULL, 0, scontext);
+#else  /*WITH_SELINUX*/
 		  add_proc (comm, pid, ppid, st.st_uid, NULL, 0);
-#endif /*FLASK_LINUX*/
+#endif /*WITH_SELINUX*/
 		else
 		  {
 		    sprintf (path, "%s/%d/cmdline", PROC_BASE, pid);
@@ -652,11 +683,11 @@ read_proc (void)
 		    (void) close (fd);
 		    if (size)
 		      buffer[size++] = 0;
-#ifdef FLASK_LINUX
-		    add_proc(comm, pid, ppid, st.st_uid, buffer, size, sid);
-#else  /*FLASK_LINUX*/
+#ifdef WITH_SELINUX
+		    add_proc(comm, pid, ppid, st.st_uid, buffer, size, scontext);
+#else  /*WITH_SELINUX*/
 		    add_proc (comm, pid, ppid, st.st_uid, buffer, size);
-#endif /*FLASK_LINUX*/
+#endif /*WITH_SELINUX*/
 		  }
 		}
 	      }
@@ -695,11 +726,11 @@ read_stdin (void)
 	cmd = comm;
       if (*cmd == '-')
 	cmd++;
-#ifdef FLASK_LINUX
-      add_proc(cmd, pid, ppid, uid, NULL, 0, -1);
-#else  /*FLASK_LINUX*/
+#ifdef WITH_SELINUX
+      add_proc(cmd, pid, ppid, uid, NULL, 0, NULL);
+#else  /*WITH_SELINUX*/
       add_proc (cmd, pid, ppid, uid, NULL, 0);
-#endif /*FLASK_LINUX*/
+#endif /*WITH_SELINUX*/
     }
 }
 
@@ -710,37 +741,37 @@ static void
 usage (void)
 {
   fprintf (stderr, _(
-    "usage: pstree [ -a ] [ -c ] [ -h | -H pid ] [ -l ] [ -n ] [ -p ] [ -u ]\n"
-    "              [ -A | -G | -U ] [ pid | user]\n"
-    "       pstree -V\n\n"
+    "Usage: pstree [ -a ] [ -c ] [ -h | -H PID ] [ -l ] [ -n ] [ -p ] [ -u ]\n"
+    "              [ -A | -G | -U ] [ PID | USER ]\n"
+    "       pstree -V\n"
+    "Display a tree of processes.\n\n"
     "    -a     show command line arguments\n"
     "    -A     use ASCII line drawing characters\n"
     "    -c     don't compact identical subtrees\n"
     "    -h     highlight current process and its ancestors\n"
-    "    -H pid highlight process \"pid\" and its ancestors\n"
+    "    -H PID highlight this process and its ancestors\n"
     "    -G     use VT100 line drawing characters\n"
     "    -l     don't truncate long lines\n"
     "    -n     sort output by PID\n"
     "    -p     show PIDs; implies -c\n"
-    "    -u     show uid transitions\n"));
-#ifdef FLASK_LINUX
-  fprintf (stderr, _(
-    "    -s     show Flask SIDs\n"
-    "    -x     show Flask security contexts\n"));
-#endif /*FLASK_LINUX*/
-  fprintf (stderr, _(
+    "    -u     show uid transitions\n"
     "    -U     use UTF-8 (Unicode) line drawing characters\n"
-    "    -V     display version information\n"
-    "    pid    start at pid, default 1 (init)\n"
-    "    user   show only trees rooted at processes of that user\n\n"));
+    "    -V     display version information\n"));
+#ifdef WITH_SELINUX
+  fprintf (stderr, _(
+    "    -Z     show SELinux security contexts\n"));
+#endif /*WITH_SELINUX*/
+  fprintf (stderr, _(
+    "    PID    start at this PID; default is 1 (init)\n"
+    "    USER   show only trees rooted at processes of this user\n\n"));
   exit (1);
 }
 
 void print_version()
 {
-  fprintf(stderr, _("pstree (psmisc) %s\n"), VERSION);
+  fprintf(stderr, _("pstree (PSmisc) %s\n"), VERSION);
   fprintf(stderr, _(
-    "Copyright (C) 1993-2002 Werner Almesberger and Craig Small\n\n"));
+    "Copyright (C) 1993-2005 Werner Almesberger and Craig Small\n\n"));
   fprintf(stderr, _(
     "PSmisc comes with ABSOLUTELY NO WARRANTY.\n"
     "This is free software, and you are welcome to redistribute it under\n"
@@ -759,7 +790,6 @@ main (int argc, char **argv)
   char termcap_area[1024];
   char *termname;
   int c;
-  char *tmpstr;
 
   if (ioctl (1, TIOCGWINSZ, &winsz) >= 0)
     if (winsz.ws_col)
@@ -774,21 +804,18 @@ main (int argc, char **argv)
   textdomain(PACKAGE);
 #endif
   
-  if ( (tmpstr = strrchr(argv[0],'/'))) {
-    tmpstr++;
-    if (strcmp(tmpstr, "pstree.x11") ==0)
-      wait_end=1;
-  }
+  if (!strcmp(__progname, "pstree.x11"))
+    wait_end=1;
 
   /*
    * Attempt to figure out a good default symbol set.  Will be overriden by
    * command-line options, if given.
    */
 
-  if (!strcmp(nl_langinfo(CODESET), "UTF-8")) {
+  if (isatty(1) && !strcmp(nl_langinfo(CODESET), "UTF-8")) {
     /* Use UTF-8 symbols if the locale's character set is UTF-8. */
     sym = &sym_utf;
-  } else if ((termname = getenv ("TERM")) && \
+  } else if (isatty (1) && (termname = getenv ("TERM")) && \
              (strlen (termname) > 0) && \
              (setupterm (NULL, 1 /* stdout */, NULL) == OK) && \
              (tigetstr ("acsc") > 0)) {
@@ -796,17 +823,20 @@ main (int argc, char **argv)
      * Failing that, if TERM is defined, a non-null value, and the terminal
      * has the VT100 graphics charset, use it.
      */
-    sym = &sym_vt100;
+	  /* problems with VT100 on some terminals, making this ascci
+	   * for now
+	   */
+    sym = &sym_ascii;
   } else {
     /* Otherwise, fall back to ASCII. */
     sym = &sym_ascii;
   }
 
-#ifdef FLASK_LINUX
-  while ((c = getopt (argc, argv, "aAcGhH:npluUVsx")) != EOF)
-#else  /*FLASK_LINUX*/
+#ifdef WITH_SELINUX
+  while ((c = getopt (argc, argv, "aAcGhH:npluUVZ")) != EOF)
+#else  /*WITH_SELINUX*/
   while ((c = getopt (argc, argv, "aAcGhH:npluUV")) != EOF)
-#endif /*FLASK_LINUX*/
+#endif /*WITH_SELINUX*/
     switch (c)
       {
       case 'a':
@@ -862,14 +892,14 @@ main (int argc, char **argv)
       case 'V':
       print_version();
 	return 0;
-#ifdef FLASK_LINUX
-      case 's':
-        show_sids = 1;
+#ifdef WITH_SELINUX
+      case 'Z':
+	if (is_selinux_enabled()>0)
+	  show_scontext = 1;
+	else
+	  fprintf(stderr, "Warning: -Z ignored. Requires anx SELinux enabled kernel\n");
         break;
-      case 'x':
-        show_scontext = 1;
-        break;
-#endif /*FLASK_LINUX*/
+#endif /*WITH_SELINUX*/
       default:
 	usage ();
       }
